/*===================================================================
This GAUSS code provides the estimation of the MSLI model described in
" A Time Varying Hidden Markov Model with Latent Information", Otranto (2007),
Statistical Modelling.

The original code was written by Chang-Jin Kim to estimate a Lam's Generalized
Hamilton Model in the state-space representation and it is available (with the data set used)  at 
http://weber.u.washington.edu/~cnelson/SSMARKOV.html.
In this version the original program is modified to insert the characteristics of the
MSLI model.

OUTPUT FILES: 
MSLI.OUT: estimates and standard errors
MSLI.DTA: data, cycle and trend
PROBS.DTA: filtered probabilities (one-step ahead and real-time) and time-varying transition probabilities
==================================================================*/

new;
library optmum,PGRAPH;
output file=MSLI.out reset;  OUTPUT OFF;


format /m1 /rd 9,4;

          load rgnp[160,2]=KIM_JE.PRN;   @ Lam's Data Set@
          qtr=rgnp[.,1];
          rgnp=rgnp[.,2];                         
          lngnp =ln(rgnp)*100;
          T0=rows(lngnp);


          yy =lngnp[2:T0,1]-lngnp[1:T0-1,1]; @N_ob is 159@
                                             @1947.2 -- 1986.4@

  T=rows(yy);  @1952.2--1984.4@

lngnp=lngnp[2:159,1];

@================= Initialize Global Variables============@


           START=2;  @1947:2.....@
           PRMTR_NO=8;   @ >>Number of parameters to be estimated. >>@
           dmnsion=2;        @ >> dimension of transition equation >>@

prmtr_in={3.5 0.0  0.5 1  0.7  -0.4 1.7 0
 };

prmtr_in=prmtr_in';

@ Maximum Likelihood Estimation @
@==================================================@

{xout,fout,gout,cout}=optmum(&lik_fcn,PRMTR_IN);

PRM_FNL=TRANS(xout);             @ Estimated  coefficients, constrained@
output on;

"==FINAL OUTPUT========================================================";

"likelihood value is ";; -1*fout;
"Estimated parameters are:";
prm_fnl';
output off;


"Calculating Hessian..... Please be patient!!!!";
                  hout0=hessp(&lik_fcn,xout);
                  hout=inv(hout0);

grdn_fnl=gradfd(&TRANS,xout);
Hsn_fnl=grdn_fnl*hout*grdn_fnl';
SD_fnl =sqrt(diag(Hsn_fnl)); @Standard errors of the estimated coefficients@

output on;
"Standard errors of parameters are:"; sd_fnl;
"===============================================================";
output off;


{CYCLE,PR_TT0,PR_TL0} =FILTER(XOUT);

TREND=LNGNP-CYCLE;

cycle1=0|cycle[1:rows(cycle)-1];
TVTP=exp(prm_fnl[1]+prm_fnl[8]*cycle1)./(1+exp(prm_fnl[1]+prm_fnl[8]*cycle1));
output file=msli.dta reset; lngnp~cycle~trend; output off;
output file=probs.dta reset; pr_tl0~pr_tt0~tvtp; output off;  @filtered probabilities@


XY(SEQA(1,1,ROWS(cycle)),cycle);   @GRAPH OF THE CYCLE@
XY(SEQA(1,1,ROWS(TREND)),TREND~LNGNP);    @GRAPH OF THE TREND@
XY(SEQA(1,1,ROWS(pr_tl0)),pr_tl0);     @GRAPH OF ONE-STEP AHEAD FILTERED PROBABILITIES@
XY(SEQA(1,1,ROWS(pr_tt0)),pr_tt0);    @GRAPH OF REAL-TIME FILTERED PROBABILITIES@
XY(SEQA(1,1,ROWS(tvtp)),tvtp);        @GRAPH OF TIME-VARYING TRANSITION PROBABILITIES@

end;

@ END OF MAIN PROGRAM @
@========================================================================@
@========================================================================@


PROC LIK_FCN(PRMTR1);

LOCAL PRMTR, PPR,QPR, PHI1,PHI2, VAR_W, MU0,MU1,R,Q,H,F,ga,a22,d22,B_LL0,B_LL1,
      VECP_LL,P_LL0,P_LL1,PROB__1,PROB__0,LIKV,J_ITER,B_TL00,B_TL01,
      B_TL10,B_TL11,P_TL00,P_TL01,P_TL10,P_TL11,F_CAST00,F_CAST01,
      F_CAST10,F_CAST11,SS00,SS01,SS10,SS11,B_TT00,B_TT01,B_TT10,
      B_TT11,P_TT00,P_TT01,P_TT10,P_TT11,PR_VL00,PR_VL01,PR_VL10,PR_VL11,
      PR_VAL,PRO_00,PRO_01,PRO_10,PRO_11,b_tt;

EXTERNAL PROC TRANS, V_PROB;


           PRMTR=TRANS(PRMTR1)|0;

           LOCATE 15,1;  PRMTR';

           PPR=exp(PRMTR[1,1])/(1+exp(prmtr[1]));    @Pr[St=1/St-1=1]@
           QPR=PRMTR[2,1];    @Pr[St=0/St-1=0]@

           PHI1=PRMTR[3,1]; PHI2=PRMTR[4,1];

           VAR_W=sqrt(PRMTR[5,1]^2);

           MU0=PRMTR[6,1];
           MU1=MU0+ PRMTR[7,1];

           R=0;
           Q=(VAR_W^2~ 0~ 0)|( 0~ 0~ 0)|( 0~ 0~ 0);
           H=(1~ -1~ 0);
         d22=prmtr[8]*exp(prmtr[1])/((1+exp(prmtr[1]))^2);
           a22=ppr;
           F =(phi1~ phi2 ~ 0)|
              ( 1  ~  0 ~ 0  )|
             (  d22~ 0~ 0);
        ga=0|0|a22;
    

/*  NOTE:  KIM'S FILTER = HAMILTON FILTER + KALMAN FILTER               */
/*  INITIALIZING THE FILTER: FOR BOTH KALMAN FILTER AND HAMILTON FILTER */

         B_LL0=0|0|a22;
         B_LL1=0|0|a22;

         VECP_LL=INV(EYE(9) - F.*.F)*vec(Q);

         P_LL0=VECP_LL[1:3,1]~VECP_LL[4:6,1]~VECP_LL[7:9,1];
         P_LL1=P_LL0;

         PROB__1=(1-QPR)/(2-PPR-QPR);   @Pr[S_0=1/Y_0], STEADY STATE PROB.@
         PROB__0=1-PROB__1  ;           @Pr[S_0=0/Y_0], STEADY STATE PROB @
 
  b_tt=PRob__0*b_ll0+PRob__1*b_ll1;

/*  START ITERATION */
/*===================================================================== */

       LIKV=0.0;
       J_ITER = 1;
       DO UNTIL J_ITER>T;


         d22=prmtr[8]*exp(prmtr[1]+prmtr[8]*b_tt[1])/((1+exp(prmtr[1]+prmtr[8]*b_tt[1]))^2);
         a22=exp(prmtr[1]+prmtr[8]*b_tt[1])/(1+exp(prmtr[1]+prmtr[8]*b_tt[1]))-d22*b_tt[1];
           F=(phi1~ phi2 ~ 0)|
              ( 1  ~  0 ~ 0  )|
             (  d22~ 0~ 0);
          ga=0|0|a22;

       /* KALMAN FILTER */
       /*==============================================================*/

             B_TL00 =ga+ F* B_LL0;  @WHEN S_{t-1}=0, S_{t}=0@
             B_TL01 = ga+F* B_LL0;  @WHEN S_{t-1}=0, S_{t}=1@
             B_TL10 = ga+F * B_LL1;  @WHEN S_{t-1}=1, S_{t}=0@
             B_TL11 = ga+F * B_LL1;  @WHEN S_{t-1}=1, S_{t}=1@
              
PPR=PRob__0*b_tl00[3]+PRob__1*b_tl11[3];

             P_TL00 = F * P_LL0 * F' + Q;
             P_TL01 = F * P_LL0 * F' + Q;
             P_TL10 = F * P_LL1 * F' + Q;
             P_TL11 = F * P_LL1 * F' + Q;

             F_CAST00= yy[J_ITER,.]- H * B_TL00- MU0;
             F_CAST01= yy[J_ITER,.]- H * B_TL01- MU1;
             F_CAST10= yy[J_ITER,.]- H * B_TL10- MU0;
             F_CAST11= yy[J_ITER,.]- H * B_TL11- MU1;


             SS00= H * P_TL00 * H' +R; @VARIANCE OF FORECAST ERROR@
             SS01= H*  P_TL01 * H' +R;
             SS10= H * P_TL10 * H' +R;
             SS11= H * P_TL11 * H' +R;

             B_TT00 = B_TL00 + (P_TL00 * H' / SS00) * F_CAST00;
             B_TT01 = B_TL01 + (P_TL01 * H' / SS01) * F_CAST01;
             B_TT10 = B_TL10 + (P_TL10 * H' / SS10) * F_CAST10;
             B_TT11 = B_TL11 + (P_TL11 * H' / SS11) * F_CAST11;

             P_TT00 = (EYE(3) - (P_TL00 * H' / SS00) * H ) * P_TL00;
             P_TT01 = (EYE(3) - (P_TL01 * H' / SS01) * H ) * P_TL01;
             P_TT10 = (EYE(3) - (P_TL10 * H' / SS10) * H ) * P_TL10;
             P_TT11 = (EYE(3) - (P_TL11 * H' / SS11) * H ) * P_TL11;


       /* HAMILTON FILTER */
       /*==============================================================*/


             PR_VL00=V_PROB(F_CAST00,SS00)*  QPR*PROB__0;    @Pr[St,Yt/Yt-1]@
             PR_VL01=V_PROB(F_CAST01,SS01)*(1-QPR)*PROB__0;
             PR_VL10=V_PROB(F_CAST10,SS10)*(1-PPR)*PROB__1;
             PR_VL11=V_PROB(F_CAST11,SS11)*    PPR*PROB__1;
                                          @CONDITIONAL DENSITY TIMES WEIGHT@

             PR_VAL=PR_VL00+PR_VL01+PR_VL10+PR_VL11;
                                  @WEIGHTED AVERAGE OF CONDITIONAL DENSITIES:
                                   f(y_t|Y_{t-1})@

             PRO_00=PR_VL00/PR_VAL;       @Pr[St,St-1/Yt]@
             PRO_01=PR_VL01/PR_VAL;
             PRO_10=PR_VL10/PR_VAL;
             PRO_11=PR_VL11/PR_VAL;

             PROB__0=PRO_00+PRO_10;       @Pr[St=0/Yt]@
             PROB__1=PRO_01+PRO_11;       @Pr[St=1/Yt]@

B_TT=B_TT00*PRO_00+B_TT01*PRO_01+B_TT10*PRO_10+B_TT11*PRO_11;

       /* COLLAPSING TERMS */
       /*==============================================================*/


              B_LL0=(PRO_00*B_TT00 + PRO_10*B_TT10)/PROB__0;
              B_LL1=(PRO_01*B_TT01 + PRO_11*B_TT11)/PROB__1;


              P_LL0=(PRO_00*(P_TT00+(B_LL0-B_TT00)*(B_LL0-B_TT00)')+
                     PRO_10*(P_TT10+(B_LL0-B_TT10)*(B_LL0-B_TT10)'))/PROB__0;

              P_LL1=(PRO_01*(P_TT01+(B_LL1-B_TT01)*(B_LL1-B_TT01)')+
                     PRO_11*(P_TT11+(B_LL1-B_TT11)*(B_LL1-B_TT11)'))/PROB__1;


              IF J_ITER < START; goto skip; endif;
              LIKV = LIKV -LN(PR_VAL);

                skip:
       J_ITER = J_ITER+1;
       ENDO;

/*  END ITERATION    */
/*===================*/


RETP(LIKV);
ENDP;



@>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>@
PROC TRANS(coef0); @ constraining values of reg. coeff.@
         local coef1,m,u,XX6,XX7;

      coef1=coef0;
      coef1[2,1]=EXP(COEF0[2,1]) ./ (1 + EXP(coef0[2,1]));

      XX6 = coef0[3,1]/(1 + abs(coef0[3,1]));
      XX7 = coef0[4,1]/(1 + abs(coef0[4,1]));

      coef1[3,1] = XX6 + XX7 ;
      coef1[4,1] =-1*XX6*XX7 ;

retp(coef1);
endp;


@>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>@
PROC V_PROB(EV, HE);      @ CALCULATES    Pr[Yt/St,Yt-1]@

         LOCAL VAL;
         VAL=(1/SQRT(2*PI*HE))*EXP(-0.5*EV*EV/HE);

RETP(VAL);
ENDP;
@>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>@


PROC (3) = FILTER(PRMTR1);

LOCAL PRMTR, PPR,QPR, PHI1,PHI2, VAR_W, MU0,MU1,R,Q,H,F,ga,a22,d22,B_LL0,B_LL1,
      VECP_LL,P_LL0,P_LL1,PROB__1,PROB__0,LIKV,J_ITER,B_TL00,B_TL01,
      B_TL10,B_TL11,P_TL00,P_TL01,P_TL10,P_TL11,F_CAST00,F_CAST01,
      F_CAST10,F_CAST11,SS00,SS01,SS10,SS11,B_TT00,B_TT01,B_TT10,
      B_TT11,P_TT00,P_TT01,P_TT10,P_TT11,PR_VL00,PR_VL01,PR_VL10,PR_VL11,
      PR_VAL,PRO_00,PRO_01,PRO_10,PRO_11,B_TT,CYCLE_M,PR_TT0M,PR_TL0M;

EXTERNAL PROC TRANS, V_PROB;


           PRMTR=TRANS(PRMTR1)|0;

           LOCATE 15,1;  PRMTR';

           PPR=exp(PRMTR[1,1])/(1+exp(prmtr[1]));    @Pr[St=1/St-1=1]@
           QPR=PRMTR[2,1];    @Pr[St=0/St-1=0]@

           PHI1=PRMTR[3,1]; PHI2=PRMTR[4,1];

           VAR_W=sqrt(PRMTR[5,1]^2);

           MU0=PRMTR[6,1];
           MU1=MU0+ PRMTR[7,1];

           R=0;
                 Q=(VAR_W^2~ 0~ 0)|( 0~ 0~ 0)|( 0~ 0~ 0);
           H=(1~ -1~ 0);
         d22=prmtr[8]*exp(prmtr[1])/((1+exp(prmtr[1]))^2);
           a22=ppr;
           F =(phi1~ phi2 ~ 0)|
              ( 1  ~  0 ~ 0  )|
             (  d22~ 0~ 0);
        ga=0|0|a22;
    

CYCLE_M=ZEROS(T,1);  @will store cyclical component: (1,1)element of B_TT@
PR_TT0M=ZEROS(T,1);  @will store Pr[S_t=0|Y_t]@
PR_TL0M=ZEROS(T,1);  @will store Pr[S_t=0|Y_{t-1}]@


/*  NOTE:  KIM'S FILTER = HAMILTON FILTER + KALMAN FILTER               */
/*  INITIALIZING THE FILTER: FOR BOTH KALMAN FILTER AND HAMILTON FILTER */

         B_LL0=0|0|a22;
         B_LL1=0|0|a22;

         VECP_LL=INV(EYE(9) - F.*.F)*vec(Q);

         P_LL0=VECP_LL[1:3,1]~VECP_LL[4:6,1]~VECP_LL[7:9,1];
         P_LL1=P_LL0;

         PROB__1=(1-QPR)/(2-PPR-QPR);   @Pr[S_0=1/Y_0], STEADY STATE PROB.@
         PROB__0=1-PROB__1  ;           @Pr[S_0=0/Y_0], STEADY STATE PROB @
b_tt=b_ll0*prob__0+b_ll1*prob__1;

/*  START ITERATION */
/*===================================================================== */

       LIKV=0.0;
       J_ITER = 1;
       DO UNTIL J_ITER>T;

         d22=prmtr[8]*exp(prmtr[1]+prmtr[8]*b_tt[1])/((1+exp(prmtr[1]+prmtr[8]*b_tt[1]))^2);
         a22=exp(prmtr[1]+prmtr[8]*b_tt[1])/(1+exp(prmtr[1]+prmtr[8]*b_tt[1]))-d22*b_tt[1];
           F=(phi1~ phi2 ~ 0)|
              ( 1  ~  0 ~ 0  )|
             (  d22~ 0~ 0);
          ga=0|0|a22;
    
PR_TL0M[J_ITER,1]= QPR*PROB__0 + (1-PPR)*PROB__1;     @Pr[St=0/Yt-1]@



       /* KALMAN FILTER */
       /*==============================================================*/

             B_TL00 = ga+F* B_LL0;  @WHEN S_{t-1}=0, S_{t}=0@
             B_TL01 = ga+F * B_LL0;  @WHEN S_{t-1}=0, S_{t}=1@
             B_TL10 = ga+F * B_LL1;  @WHEN S_{t-1}=1, S_{t}=0@
             B_TL11 = ga+F * B_LL1;  @WHEN S_{t-1}=1, S_{t}=1@

PPR=PRob__0*b_tl00[3]+PRob__1*b_tl11[3];


             P_TL00 = F* P_LL0 * F' + Q;
             P_TL01 = F * P_LL0 * F' + Q;
             P_TL10 = F * P_LL1 * F' + Q;
             P_TL11 = F * P_LL1 * F' + Q;

             F_CAST00= yy[J_ITER,.]- H * B_TL00- MU0;
             F_CAST01= yy[J_ITER,.]- H * B_TL01- MU1;
             F_CAST10= yy[J_ITER,.]- H * B_TL10- MU0;
             F_CAST11= yy[J_ITER,.]- H * B_TL11- MU1;


             SS00= H * P_TL00 * H' +R; @VARIANCE OF FORECAST ERROR@
             SS01= H*  P_TL01 * H' +R;
             SS10= H * P_TL10 * H' +R;
             SS11= H * P_TL11 * H' +R;

             B_TT00 = B_TL00 + (P_TL00 * H' / SS00) * F_CAST00;
             B_TT01 = B_TL01 + (P_TL01 * H' / SS01) * F_CAST01;
             B_TT10 = B_TL10 + (P_TL10 * H' / SS10) * F_CAST10;
             B_TT11 = B_TL11 + (P_TL11 * H' / SS11) * F_CAST11;

             P_TT00 = (EYE(3) - (P_TL00 * H' / SS00) * H ) * P_TL00;
             P_TT01 = (EYE(3) - (P_TL01 * H' / SS01) * H ) * P_TL01;
             P_TT10 = (EYE(3) - (P_TL10 * H' / SS10) * H ) * P_TL10;
             P_TT11 = (EYE(3) - (P_TL11 * H' / SS11) * H ) * P_TL11;


       /* HAMILTON FILTER */
       /*==============================================================*/


             PR_VL00=V_PROB(F_CAST00,SS00)*  QPR*PROB__0;    @Pr[St,Yt/Yt-1]@
             PR_VL01=V_PROB(F_CAST01,SS01)*(1-QPR)*PROB__0;
             PR_VL10=V_PROB(F_CAST10,SS10)*(1-PPR)*PROB__1;
             PR_VL11=V_PROB(F_CAST11,SS11)*    PPR*PROB__1;
                                          @CONDITIONAL DENSITY TIMES WEIGHT@

             PR_VAL=PR_VL00+PR_VL01+PR_VL10+PR_VL11;
                                  @WEIGHTED AVERAGE OF CONDITIONAL DENSITIES:
                                   f(y_t|Y_{t-1})@

             PRO_00=PR_VL00/PR_VAL;       @Pr[St,St-1/Yt]@
             PRO_01=PR_VL01/PR_VAL;
             PRO_10=PR_VL10/PR_VAL;
             PRO_11=PR_VL11/PR_VAL;

B_TT=B_TT00*PRO_00+B_TT01*PRO_01+B_TT10*PRO_10+B_TT11*PRO_11;
CYCLE_M[J_ITER,1]=B_TT[1,1];

             PROB__0=PRO_00+PRO_10;       @Pr[St=0/Yt]@
             PROB__1=PRO_01+PRO_11;       @Pr[St=1/Yt]@

PR_TT0M[J_ITER,1]=PROB__0;

       /* COLLAPSING TERMS */
       /*==============================================================*/


              B_LL0=(PRO_00*B_TT00 + PRO_10*B_TT10)/PROB__0;
              B_LL1=(PRO_01*B_TT01 + PRO_11*B_TT11)/PROB__1;


              P_LL0=(PRO_00*(P_TT00+(B_LL0-B_TT00)*(B_LL0-B_TT00)')+
                     PRO_10*(P_TT10+(B_LL0-B_TT10)*(B_LL0-B_TT10)'))/PROB__0;

              P_LL1=(PRO_01*(P_TT01+(B_LL1-B_TT01)*(B_LL1-B_TT01)')+
                     PRO_11*(P_TT11+(B_LL1-B_TT11)*(B_LL1-B_TT11)'))/PROB__1;


              IF J_ITER < START; goto skip; endif;
              LIKV = LIKV -LN(PR_VAL);

                skip:
       J_ITER = J_ITER+1;
       ENDO;

/*  END ITERATION    */
/*===================*/


RETP(CYCLE_M[START:T],PR_TT0M[START:T],PR_TL0M[START:T]);
ENDP;


@==========================================================@



